import rdflib
import rdfextras
rdfextras.registerplugins()
import sys
if sys.version_info[0] < 3:
    from StringIO import StringIO
else:
    from io import StringIO
from pandas import DataFrame

'''
sparql.py: part of the nidmviewer package
Sparql queries

'''

def do_query(ttl_file,query,rdf_format="turtle",serialize_format="csv",output_df=True):
    g = rdflib.Graph()
    g.parse(ttl_file,format=rdf_format)
    result = g.query(query)   
    if result is None:
        print "No results matching query."
    else:
        print "Found results matching query."
        result = result.serialize(format=serialize_format)    
        if output_df == True:
            result = StringIO(result)
            return DataFrame.from_csv(result,sep=",")
    return result

def get_coordinates(ttl_file):
    query = """
    SELECT DISTINCT ?name ?coordinate ?z_score ?peak_name ?pvalue_uncorrected
    WHERE {?coord a nidm:NIDM_0000015 ;
           rdfs:label ?name ;
           nidm:NIDM_0000086 ?coordinate .
       ?peak prov:atLocation ?coord ;
           nidm:NIDM_0000092 ?z_score ;
           rdfs:label ?peak_name ;
           nidm:NIDM_0000116 ?pvalue_uncorrected .}
     ORDER BY ?name
    """
    return do_query(ttl_file,query)


def get_coordinates_and_maps(ttl_file):
    query = """
            PREFIX nidm: <http://purl.org/nidash/nidm#>
            PREFIX prov: <http://www.w3.org/ns/prov#>
            prefix nfo: <http://www.semanticdesktop.org/ontologies/2007/03/22/nfo#>
            prefix spm: <http://purl.org/nidash/spm#>
            prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>
            prefix peak: <http://purl.org/nidash/nidm#NIDM_0000062>
            prefix significant_cluster: <http://purl.org/nidash/nidm#NIDM_0000070>
            prefix coordinate: <http://purl.org/nidash/nidm#NIDM_0000086>
            prefix equivalent_zstatistic: <http://purl.org/nidash/nidm#NIDM_0000092>
            prefix pvalue_fwer: <http://purl.org/nidash/nidm#NIDM_0000115>
            prefix pvalue_uncorrected: <http://purl.org/nidash/nidm#NIDM_0000116>
            prefix statistic_map: <http://purl.org/nidash/nidm#NIDM_0000076>
            prefix statistic_type: <http://purl.org/nidash/nidm#NIDM_0000123>
            prefix nidm_ExcursionSetMap: <http://purl.org/nidash/nidm#NIDM_0000025>
            SELECT DISTINCT ?statmap ?excsetmap_location ?statmap_type ?z_score 
            ?pvalue_uncorrected ?coordinate_id ?coord_name ?coordinate ?exc_set
            WHERE {
            ?statmap a statistic_map: ;
                statistic_type: ?statmap_type .
            ?exc_set a nidm_ExcursionSetMap: ;
                prov:wasGeneratedBy/prov:used ?statmap ;
                prov:atLocation ?excsetmap_location .
            OPTIONAL {
            ?peak prov:wasDerivedFrom/prov:wasDerivedFrom/prov:wasGeneratedBy/prov:used ?statmap ;
                prov:atLocation ?coord ;
                equivalent_zstatistic: ?z_score ;
                pvalue_uncorrected: ?pvalue_uncorrected ;
                prov:atLocation ?coordinate_id .
            ?coordinate_id rdfs:label ?coord_name ;
                coordinate: ?coordinate .
                }
            }
            """
    return do_query(ttl_file,query)



'''
templates.py: part of the nidmviewer package
Functions to work with html templates

'''

from nidmviewer.utils import get_package_dir
import pandas
import os
import re

def get_template(html_name):
    return read_template(html_name)

# Add code string to end of template
def add_javascript_function(function_code,template):
    template.append("<script>\n%s\n</script>" % (function_code))
    return template

# Remove scripts (css or js) from html_snippet
def remove_resources(html_snippet,script_names):
    expression = re.compile("|".join(script_names))
    filtered_template = [x for x in html_snippet if not expression.search(x)]
    return filtered_template

def save_template(html_snippet,output_file):
    filey = open(output_file,"wb")
    filey.writelines(html_snippet)
    filey.close()

def read_template(html_name):
    ppwd = get_package_dir()  
    html_name = html_name + ".html"
    template_file = os.path.abspath(os.path.join(ppwd,'template', html_name))
    filey = open(template_file,"r")
    template = "".join(filey.readlines())
    filey.close()
    return template

def add_string(tag,substitution,template):
    template = template.replace(tag,substitution)
    return template

'''Get an image by name in the img directory'''
def get_image(image_name):
    ppwd = get_package_dir()  
    return os.path.join(ppwd,'img', image_name)



'''
browser.py: part of nidmviewer package
Functions to visualize in browser

'''
from nidmviewer.utils import make_tmp_folder
import SimpleHTTPServer
import SocketServer
import webbrowser
import shutil
import os

'''
View code in temporary browser!
html_snippet is the template code with images subbed
copy_list is a dictionary, 
  with {nidm:{brainmap1_file:temp1_file,..brainmapN_file:tempN_file}}
'''
def view(html_snippet,copy_list,port):
    with make_tmp_folder() as tmp_dir:  
        # First copy all brain maps
        for real_path,temp_path in copy_list.iteritems():
            real_path = os.path.abspath(real_path.replace("file://",""))
            shutil.copy(real_path,"%s/%s" %(tmp_dir,temp_path))
        # Now write template to temporary file
        tmp_file = "%s/pycompare.html" %(tmp_dir)
        # Change directory and start a web server
        os.chdir(tmp_dir)
        print os.getcwd()
        write_file(html_snippet,tmp_file)
        tmp_file_base = os.path.basename(tmp_file)
        if port!=None:
            httpd = run_webserver(html_page="%s" %(tmp_file_base),port=port)
        else:
            httpd = run_webserver(html_page="%s" %(tmp_file_base))
        return httpd


'''Internal view function'''
def internal_view(html_snippet,tmp_file):
    url = 'file://%s' %(tmp_file)
    write_file(html_snippet,tmp_file)
    webbrowser.open_new_tab(url)
    raw_input("Press Enter to finish...")

def write_file(html_snippet,tmp_file):
    html_file = open(tmp_file,'wb')
    html_file.writelines(html_snippet)
    html_file.close()

'''Web server (for Papaya Viewer in QA report'''
def run_webserver(port=8088,html_page="index.html"):
    Handler = SimpleHTTPServer.SimpleHTTPRequestHandler
    httpd = SocketServer.TCPServer(("", port), Handler)
    print "Serving pybraincompare at port", port
    webbrowser.open("http://localhost:%s/%s" %(port,html_page))
    httpd.serve_forever()
    return httpd

"""Get svg html from matplotlib figures (eg, glass brain images)"""
def get_svg_html(mpl_figures):
    svg_images = []
    with make_tmp_folder() as tmp_dir:  
        for fig in mpl_figures:
            tmp_svg = "%s/mplfig.svg" %(tmp_dir)
            fig.savefig(tmp_svg)
            fig_data = open(tmp_svg,"rb").readlines()
            svg_images.append(fig_data)
    return svg_images



#!/usr/bin/env python

'''
script.py: part of nidmviewer package
Functions to visualize in browser

'''

from nidmviewer.viewer import generate
from glob import glob
import argparse
import sys
import os

def main():
    parser = argparse.ArgumentParser(
    description="command line or server tool to view or compare nidm results.")
    parser.add_argument("ttl", help="List of comma separated ttl files to parse.", type=str)
    parser.add_argument("--base", help="base image (standard brain map) to use for the viewer background.",type=str)
    parser.add_argument("--port", help="PORT to use to serve nidmviewer (default 8088).",type=int)
    parser.add_argument("--columns_to_remove", help="Comma separated list of columns to remove from viewer.",type=str,default="statmap_filename,excsetmap_location,statmap_type")
 
    try:
        args = parser.parse_args()
    except:
        parser.print_help()
        sys.exit(0)

    print "Starting up the nidmviewer!"

    nidm_files = args.ttl.split(",")
    if args.columns_to_remove != None:
        args.columns_to_remove = [x.strip(" ") for x in args.columns_to_remove.split(",")]
   
    nidm_files = [os.path.abspath(f) for f in nidm_files]
    if args.base != None:
        standard_brain = os.path.abspath(args.base)
    else:
        standard_brain = None

    httpd = generate(ttl_files=nidm_files,
                   base_image=standard_brain,
                   columns_to_remove=args.columns_to_remove,
                   view_in_browser=True,
                   port=args.port)

if __name__ == '__main__':
    main()



"""
convert

Functions to convert/parse output from nidm sparql queries

"""

from nidmviewer.utils import read_file_lines
from numpy import isnan, float64
import pandas 

def parse_coordinates(coordinates):
    '''parse_coordinates
    convert a list of xyz strings in format '[x,y,z]' to separate variables
    This is what we get from the sparql query
    Parameters
    ==========
    coordinates: list
        a list of xyz coordinate strings, each a list in a string '[x,y,z]'
    Returns
    =======
    coordinate_df: pandas dataframe
    columns are x,y,z
    '''
    count=0
    coordinate_df = pandas.DataFrame(columns=["x","y","z"])
    for coordinate in coordinates:
        if isinstance(coordinate, str):
            coordinate_df.loc[count] = [x.strip() for x in coordinate.strip("]").strip("[").split(",")]
        else:
            coordinate_df.loc[count] = [None,None,None]
        count+=1
    return coordinate_df

def getjson(nidm_file,format="n3"):
    g = Graph().parse(nidm_file, format=format)
    return json.loads(g.serialize(format='json-ld', indent=4))

def prettyjson(nidm_file,format="n3"):
    g = Graph().parse(nidm_file, format=format)
    return g.serialize(format='json-ld', indent=4)




'''
viewer.py: part of the nidmviewer package

'''
from nidmviewer.utils import strip_url, get_random_name, get_extension, \
 download_file, get_standard_brain, is_empty, get_images
from nidmviewer.templates import get_template, add_string, save_template, remove_resources
from nidmviewer.sparql import get_coordinates_and_maps
from nidmviewer.convert import parse_coordinates
from nidmviewer.browser import view
import pandas
import numpy
import os
import sys


def generate(ttl_files,base_image=None,retrieve=False,view_in_browser=False,columns_to_remove=None,
             template_choice="index",port=None,remove_scripts=None,button_text="BRAIN",check_empty=False):
    '''generate
    will generate a nidmviewer to run locally or to embed into webserver
    Parameters
    ==========
    ttl_files: str or list 
        one or more turtle files to add to the viewer. Images in the files should be available
        at the specified URL. 
    retrieve: boolean
        If set to False, the images are assumed to be on the same server, and will be 
        served with the given URL. If retrieve is set to True, the images will be retrieved
        first and stored in a temporary directory.
    base_image: str
        The base image to use for the viewer. Not specifying a base_image will
        yield a black background.
    view_in_browser: boolean
        open a temporary web browser (to run locally). If True, images will be copied
        to a temp folder. If False, image_paths must be relative to web server. File names 
        should be unique.
    columns_to_remove: additional columns to remove. If none, default columns of "coordinate_id"
        "statmap_type" and "exc_set" will be removed.
    port: int
        port to serve nidmviewer, if view_in_browser==True
    remove_scripts: list
        one or more script or button tags to remove from the template. Options include
        JQUERY BOOTSTRAPJS BOOTSTRAPCSS PAPAYACSS PAPAYAJS NIDMSELECTBUTTON
    button_text: str
        Text string for the button to select a brain image. Default is "BRAIN"
    check_empty: boolean - check for empty images or not. Will result in error if nidm paths
               are URLS.
    '''

    # Check inputs and prepare template        
    ttl_files = check_inputs(ttl_files)
    template = get_template(template_choice)  

    if remove_scripts != None:
        if isinstance(remove_scripts,str):
            remove_scripts = [remove_scripts]
        template = template.split("\n")
        resources = ["TAG_%s_TAG" %x for x in remove_scripts]
        template = remove_resources(template,resources)
        template = "\n".join(template)

    # Parse each nidm file
    peaks = parse_nidm(ttl_files)

    # Convert coordinates from '[x,y,z]' to [x],[y],[z]
    for nidm,peak in peaks.iteritems():
        coordinates = parse_coordinates(peak.coordinate.tolist())
        peak = peak.drop("coordinate",axis=1)
        if coordinates.shape[0] > 0:
            peak["x"] = coordinates["x"].tolist()
            peak["y"] = coordinates["y"].tolist()
            peak["z"] = coordinates["z"].tolist()
        peaks[nidm] = peak

    # Grab the column names, it could be different for each ttl
    column_names = get_column_names(peaks)

    # if the user wants to remove columns
    columns = ["coordinate_id","exc_set","statmap_type"]
    if columns_to_remove != None:
        if isinstance(columns_to_remove,str):
            columns_to_remove = [columns_to_remove]
        columns = columns_to_remove + columns
    column_names = remove_columns(column_names,columns)

    # We want pandas df in the format of dict/json strings for javascript embed
    for nidm,peak in peaks.iteritems():
        peaks[nidm] = to_dictionary(peak,strings=True)    
 
    # Retrieve nifti files, if necessary
    peaks = retrieve_nifti(peaks,retrieve,"excsetmap_location")

    template = add_string("[SUB_FILELOCATIONKEY_SUB]","excsetmap_location",template)
    template = add_string("[SUB_FILENAMEKEY_SUB]","statmap",template)
    template = add_string("[SUB_COLUMNS_SUB]",str(column_names),template)
    template = add_string("[SUB_BUTTONTEXT_SUB]",button_text,template)

    # If excursion set is empty (regardless of peaks listed in the .ttl file) show "No suprathreshold voxels" instead of the table and the nifiti viewer. (this will happen if the analysis did not return any statistically significant results)
    empty_images = dict()

    if view_in_browser==True:
        peaks,copy_list = generate_temp(peaks,"excsetmap_location")

        for exc_set_file,image_name in copy_list.iteritems():
            if check_empty == True:
                empty_images[image_name] = is_empty(exc_set_file)
            else:
                empty_images[image_name] = 0

        if base_image == None:
            base_image = get_standard_brain(load=False)
            dummy_peak = {base_image: [{"excsetmap_location":base_image}]}
            junk,base_copy = generate_temp(dummy_peak,"excsetmap_location")
            copy_list.update(base_copy)
            base_image = base_copy.values()[0]
        
        template = add_string("[SUB_EMPTY_SUB]",str(empty_images),template)
        template = add_string("[SUB_PEAKS_SUB]",str(peaks),template)
        template = add_string("[SUB_BASEIMAGE_SUB]",str(base_image),template)
        view(template,copy_list,port)

    else:
        if base_image == None:
            base_image = get_standard_brain(load=False)

        image_files = get_images(peaks,"excsetmap_location")
        for image_file in image_files:
            if check_empty == True:
                empty_images[image_file] = is_empty(image_file)
            else:
                empty_images[image_file] = 0

        template = add_string("[SUB_EMPTY_SUB]",str(empty_images),template)
        template = add_string("[SUB_PEAKS_SUB]",str(peaks),template)
        template = add_string("[SUB_BASEIMAGE_SUB]",str(base_image),template)
        return template


def parse_nidm(ttl_files):
    '''parse_nidm
        Extract brainmaps and coordinates from ttl files
    Parameters:
    ==========
    ttl_files: list
        list of full paths to ttl files
    Returns
    =======
    peaks: dict
        dict of pandas data frames, one for each ttl_file, with columns
        coordinate, z, peak_name, pvalue_uncorrected
    maps: dict
        dict of pandas data frames, one for each ttl_file, with columns
        filename and location for all brain maps specified in ttl.
    '''
    peaks = dict()
    for n in range(len(ttl_files)):
        ttl_file = os.path.abspath(ttl_files[n])
        df = get_coordinates_and_maps(ttl_file)
        peaks[ttl_file] = df
    return peaks


def to_dictionary(df,orient="records",strip_columns=False,strings=False):
    '''to_dictionary: 
       Convert a pandas dataframe into the string of a json/dict to 
       embed into page
       Parameters
       ==========
       df: pandas data frame
           data frame to convert
       orient: str
           orientation to convert with (default is "records")
       strip_columns: boolean
           if true, will return df.to_dict(orient=orient).values()
           default is False
       strings: boolean
           True will convert all columns to strings.
       '''
    if strings:
        for column in df.columns:
            df[column] = df[column].astype(str)
    if strip_columns:
        return df.to_dict().values() 
    else:
        return df.to_dict(orient=orient)


def retrieve_nifti(peaks,retrieve,location_key):
    '''retrieve_nifti
    Download the image to a temporary folder if the user needs to
    retrieve it. Otherwise, return file
    Parameters
    ==========
    peaks: dict
        dictionary (key, is ttl_file, and value, is dictionary of {filename:fullpath}
        for all brainmaps extracted from the ttl files
    retrieve: boolean
        if True, will download brainmaps to temporary folder first. If false, encodes
        path in utf-8 for rendering in javascript
    location_key: str
        key to look up file name in peaks dictionary
    '''
    # Note: retrieve = True has not been tested!
    updated_peaks = dict()
    for nidm,entries in peaks.iteritems():
        if retrieve:
            for e in range(len(entries)):
                if location_key in entries[e]:
                    brainmap = entries[e][location_key]
                    image_ext = get_extension(brainmap)
                    temp_path = get_random_name()
                    temp_dir = tempfile.mkdtemp()
                    temp_image_path = "%s/%s.%s" %(temp_dir,temp_path,image_ext)
                    if download_file(brainmap,temp_image_path):
                        entries[e][location_key] = temp_image_path
        updated_peaks[nidm] = entries
    return updated_peaks


def get_column_names(peaks):
    column_names = dict()
    for nidm,df in peaks.iteritems():
        column_names[nidm] = df.columns.tolist()
    return column_names


def generate_temp(peaks,location_key):
    '''generate_temp
    generate a lookup of temporary files
    Parameters
    ==========
    peaks: dict
       data structure from get_coordinates_and_peaks
    location_key: str
       key in peaks data structure for file paths
    Returns
    =======
    peaks: dict
        (key is ttl file, equivalent to peaks, but old location_key path is replaced
        with path to temporary directory
    copy_list: dict
        keys are current paths, values are temporary file names corresponding to
        fullpath in new_nifti_files[ttl_file] dictionary. This is used to copy
        images into the temporary directory with the correct names. 
    '''
    copy_list = dict()
    for nidm,entries in peaks.iteritems():
        nidm_directory = os.path.dirname(nidm)
        for e in range(len(entries)):
            if location_key in entries[e]:
                brainmap = entries[e][location_key]
                brainmap_base = os.path.basename(brainmap)
                if "%s/%s" %(nidm_directory,brainmap_base) not in copy_list:
                    image_ext = get_extension(brainmap)
                    temp_path = get_random_name()
                    temp_image_path = "%s.%s" %(temp_path,image_ext)
                    copy_list["%s/%s" %(nidm_directory,brainmap_base)] = temp_image_path

    updated_peaks = dict()

    for nidm,entries in peaks.iteritems():
        nidm_directory = os.path.dirname(nidm)
        to_keep = []
        for e in range(len(entries)):
            if location_key in entries[e]:
                brainmap = entries[e][location_key]
                brainmap_base = os.path.basename(brainmap)
                entries[e][location_key] = copy_list["%s/%s" %(nidm_directory,brainmap_base)]
                if "x" in entries[e]:
                    if entries[e]["x"] != "nan":
                        to_keep.append(e)

        # Remove coordinates that are nan from the data frame
        entries = [entries[e] for e in range(len(entries)) if e in to_keep]
        updated_peaks[nidm] = entries

    return updated_peaks,copy_list 


def check_inputs(ttl_files):
    """check_input.
    will return a list of ttl files, more advanced
    functionality can be added if needed (validation, etc)
    Parameters
    ==========
    ttl_files: string or list
        full paths to ttl files   
    Returns
    =======
        list of ttl files
    """
    if isinstance(ttl_files,str): ttl_files = [ttl_files]
    return ttl_files

def remove_columns(columns,columns_to_remove):
    new_columns = dict()
    for nidm,column_names in columns.iteritems():
        new_columns[nidm] = [x for x in column_names if x not in columns_to_remove]        
    return new_columns



'''
utils.py: part of the nidmviewer package
Functions to work with html templates

'''

import numpy as np
import contextlib
import tempfile
import nibabel
import zipfile
import urllib
import shutil
import string
import numpy
import random
import __init__
import os


# Split a url to just get base
def strip_url(url,encode=True):
    if encode:
        return url.split("/")[-1].encode("utf-8")
    else:
        return url.split("/")[-1]

# Get the directory of the package
def get_package_dir():
    return os.path.abspath(os.path.join(os.path.dirname(__init__.__file__)))

# Make directory
def make_dir(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)

# Unzip static files to temporary directory
def unzip(source,dest_dir):
    with zipfile.ZipFile(source, "r") as z:
        z.extractall(dest_dir)

# Make temporary directory
@contextlib.contextmanager
def make_tmp_folder():
    temp_dir = tempfile.mkdtemp()
    yield temp_dir
    shutil.rmtree(temp_dir)

# Reading Files
def read_file_lines(file_name):
    filey = open(file_name,'r')
    file_contents = filey.readlines()
    filey.close()
    return file_contents


def download_file(src,dest):
    try:
        requester = urllib.URLopener()
        requester.retrieve(src, dest)
        return True
    except:
        print "Cannot download %s" %(src)
        return False


# Filenames
def get_name(path):
    return os.path.split(path)[1].split(".")[0]

def get_random_name(length=6,chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(length))

# Generate new png paths to correspond to nifti filenames
def make_png_paths(nifti_files):
    image_paths = []
    for i in range(0,len(mr_files)):
        image = mr_files[i]
        tmp_svg = "%s/tmpbrain%s.png" %(tmp_dir,i)
        make_glassbrain_image(image,tmp_svg)
        image_paths.append(tmp_svg)
    return image_paths

def get_extension(path):
    fileparts =  os.path.basename(path).split(".")
    fileparts.pop(0)
    return ".".join(fileparts)

# Get unique values in a list of lists
def unwrap_list_unique(list_of_lists):
    uniques = []
    for listy in list_of_lists:
        uniques = uniques + [item for item in listy]
    uniques = list(np.unique(uniques))
    return uniques

# Brain image templates
def get_standard_brain(load=True):
    mr_directory = get_package_dir()
    brain = "%s/data/MNI152_T1_2mm_brain.nii.gz" %(mr_directory)
    if load:
        return nib.load(brain)
    else:
        return brain

# Check if nifti is empty
def is_empty(nii_file):
    nii = nibabel.load(nii_file)
    data = nii.get_data()
    data = numpy.nan_to_num(data)
    if numpy.count_nonzero(data) == 0:
        return 1
    return 0

def get_images(peaks,location_key):
    '''get_images returns unique images for a location key from
    a peaks table
    '''
    image_list = []
    for nidm,entries in peaks.iteritems():
        nidm_directory = os.path.dirname(nidm)
        for e in range(len(entries)):
            if location_key in entries[e]:
                brainmap = entries[e][location_key]
                if brainmap not in image_list:
                    image_list.append(brainmap)
    return image_list



from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the relevant file
with codecs.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get requirements from requirements.txt file
reqs = [line.strip() for line in open('requirements.txt').readlines()]
requirements = list(filter(None, reqs))

setup(
    # Application name:
    name="nidmviewer",

    # Version number (initial):
    version="0.1.2",

    # Application author details:
    author="Vanessa Sochat",
    author_email="vsochat@stanford.edu",

    # Packages
    packages=find_packages(),

    # Data
    package_data = {'nidmviewer.template':['*.html','*.zip','*.js','*.css'],
                    'nidmviewer.data':['*.nii.gz']},

    # Details
    url="http://www.github.com/vsoch/nidmviewer",

    license="LICENSE.txt",
    description="command line or server tool to view or compare nidm results.",
    long_description=long_description,
    keywords='nidm nidm-results brain imaging neuroimaging',

    install_requires = requirements,

    entry_points = {
        'console_scripts': [
            'nidmviewer=nidmviewer.scripts:main',
        ],
    },
)
